{
@abstract(provides HTML doc generator object)
@author(Erwin Scheuch-Heilig (ScheuchHeilig@t-online.de))
@author(Marco Schmidt (marcoschmidt@geocities.com))
@created(1 Sep 1998)
@lastmod(10 Apr 2000)

Implements an object to generate HTML documentation, overriding many of
@link(TDocGenerator)'s virtual methods.
}

unit Html;

{$I platform.inc}

interface

uses
  AppInfo,
  Arrays,
  FileStre,
  GenDoc,
  Items,
  Numbers,
  Objects,
  Parsing,
  Streams,
  Texts,
  Time;

type
  THTMLColorIndex = (
    HTML_BACKGROUND,
    HTML_TEXT,
    HTML_LINK,
    HTML_VLINK,
    HTML_ALINK,
    HTML_TABLE_BACKGROUND,
    HTML_TABLE_TEXT);

type
  { pointer to @link(THTMLDocGenerator) }
  PHTMLDocGenerator = ^THTMLDocGenerator;
  { @abstract(generates HTML documentation)
    Extends @link(TDocGenerator) and overwrites many of its methods to generate
    output in HTML (HyperText Markup Language) format.
    This type of output is well suited to be read with a web browser at the
    computer, as a reference manual that does not have to be printed.
    For printed output, use @link(Tex.TTexDocGenerator). }
  THTMLDocGenerator = object(TDocGenerator)
    { Makes a string look like a coded string, i.e. <CODE>TheString</CODE>
      in Html. }
    function CodeString(const s: string): string; virtual;
    { Converts C to HTML, simply copying normal characters and converting
      special characters like the ampersand to escape sequences. }
    function ConvertChar(c: Char): string; virtual;
    { Returns a link to an anchor within a document. HTML simply concatenates
      the strings with a "#" character between them. }
    function CreateLink(Item: PItem): string; virtual;
    { Creates a valid HTML link, starting with an anchor that points to Link,
      encapsulating the text ItemName in it. }
    function CreateReferencedLink(ItemName, Link: string): string; virtual;
    function ExistsFullPath(s: string): Boolean;
    { Returns HTML file extension ".htm". }
    function GetFileExtension: string; virtual;
    { Writes information on doc generator to current output stream,
      including link to pasdoc homepage. }
    procedure WriteAppInfo;
    { Writes authors to output, at heading level HL. Will not write anything
      if collection of authors is not assigned or empty. }
    procedure WriteAuthors(HL: Byte; Authors: PStringCollection);
    { Writes a single class, interface or object CIO to output, at heading
      level HL. }
    procedure WriteCIO(HL: Byte; var CIO: TCIO); virtual;
    { Calls @link(WriteCIO) with each element in the argument collection C,
      using heading level HL. }
    procedure WriteCIOs(HL: Byte; c: PItemCollection); virtual;
    procedure WriteCIOSummary(HL: Byte; c: PItemCollection); virtual;
    procedure WriteCodeWithLinks(p: PItem; s: string);
    { Writes a comment S to output file. }
    procedure WriteComment(const s: string); virtual;
    { Writes dates Created and LastMod at heading level HL to output
      (if at least one the two has a value assigned). }
    procedure WriteDates(HL: Byte; Created, LastMod: PString); virtual;
    { The method that does everything - writes documentation for all units
      and creates overview files. }
    procedure WriteDocumentation; virtual;
    { Writes the beginning of the HTML document, including opening HTML element,
      a complete HEAD element and an opening BODY element.
      See @link(WriteEndOfDocument). }
    procedure WriteDocumentHeadline; virtual;
    { Writes an empty table cell, '&nbsp;'. }
    procedure WriteEmptyCell;

    procedure WriteEndOfCode; virtual;
    { Writes the end of an HTML anchor, '</A>'. }
    procedure WriteEndOfAnchor;
    { See @link(WriteDocumentHeadline). }
    procedure WriteEndOfDocument;
    procedure WriteEndOfLink;
    { Finishes an HTML paragraph element by writing a closing P tag. }
    procedure WriteEndOfParagraph;
    { Finishes an HTML table cell by writing a closing TD tag. }
    procedure WriteEndOfTableCell;
    { Finishes an HTML table by writing a closing TABLE tag. }
    procedure WriteEndOfTable;
    { Finishes an HTML table row by writing a closing TR tag. }
    procedure WriteEndOfTableRow;
    procedure WriteFooter;
    { Writes information on functions and procedures or methods of a unit or
      class, interface or object to output.
      If argument Methods is true, they will be considered methods of a class,
      interface or object, otherwise they're considered functions or procedures
      of a unit.
      The functions are stored in the FuncsProcs argument. }
    procedure WriteFuncsProcs(HL: Byte; Methods: Boolean; FuncsProcs: PMethodCollection); virtual;
    { Writes heading S to output, at heading level I.
      For HTML, only levels 1 to 6 are valid, so that values smaller
      than 1 will be set to 1 and arguments larger than 6 are set to 6.
      The string S will then be enclosed in an element from H1 to H6,
      according to the level. }
    procedure WriteHeading(i: Byte; const s: string); virtual;
    procedure WriteItemDescription(Item: PItem);
    procedure WriteItemDetailedDescription(Item: PItem);
    procedure WriteItems(Order: Byte; Heading: string; i: PItemCollection); virtual;
    procedure WriteOverviewFiles;
    procedure WriteParagraph(HL: Byte; s: string; t: PText);
    procedure WriteProperties(HL: Byte; p: PPropertyCollection); virtual;
    { Writes an opening A element, including a name attribute given by the
      argument. }
    procedure WriteStartOfAnchor(const Name: string);
    procedure WriteStartOfCode; virtual;
    procedure WriteStartOfDocument(Name: string); virtual;
    procedure WriteStartOfLink(const Name: string);
    { Starts an HTML paragraph element by writing an opening P tag. }
    procedure WriteStartOfParagraph;
    procedure WriteStartOfTableCell;
    procedure WriteStartOfTable1Column(t: string);
    procedure WriteStartOfTable2Columns(t1, t2: string);
    procedure WriteStartOfTable3Columns(t1, t2, T3: string);
    procedure WriteStartOfTableRow;
    procedure WriteUnit(HL: Byte; U: PUnit); virtual;
    procedure WriteUnitDescription(HL: Byte; d: PText); virtual;
    { Creates an output stream that lists up all units and short descriptions. }
    procedure WriteUnitOverviewFile;
  end;

const
  { background color of a table header row; a light gray slightly darker
    than the light gray of @link(HTML_HEADER_BACKGROUND_COLOR)  }
  HTML_ROW_BACKGROUND_COLOR: string[6] = 'efefef';
  { background color of a normal table row; a light gray slightly lighter
    than the light gray of @link(HTML_ROW_BACKGROUND_COLOR) }
  HTML_HEADER_BACKGROUND_COLOR: string[6] = 'e0e0e0';

  HTML_TABLE_CELLPADDING = '4';
  HTML_TABLE_CELLSPACING = '2';

implementation

uses
  SysUtils,

  Msg;
{
html things to be customized
- standard background color (white)
- background color for table headings and overview list at the top of each file (light gray)
- background color for normal table cells (light gray, slightly lighter than the above)
- standard foreground color (black)
- unused link color (blue)
- used link color (purple)
- link color while being clicked on (red)
- normal font (Times Roman)
- heading font (Helvetica)
- code font (Courier)
}

function THTMLDocGenerator.CodeString(const s: string): string;
begin
  Result := '<CODE>' + s + '</CODE>';
end;

function THTMLDocGenerator.ConvertChar(c: Char): string;
begin
  case c of
    { characters already in use by HTML }
    // '<': Result := '&lt;';              { less than }
    // '>': Result := '&gt;';              { greater than }
    // '&': Result := '&amp;';             { ampersand }
    // '"': Result := '&quot;';            { quote }
    { some German special characters }
    #$E4: Result := '&auml;'; {  (small Umlaut a) }
    #$F6: Result := '&ouml;'; {  (small Umlaut o) }
    #$FC: Result := '&uuml;'; {  (small Umlaut u) }
    #$C4: Result := '&Auml;'; {  (large Umlaut A) }
    #$D6: Result := '&Ouml;'; {  (large Umlaut O) }
    #$DC: Result := '&Uuml;'; { U (large Umlaut U) }
    #$DF: Result := '&szlig;'; {  (sz) }
    { more special characters to be added here }
    { otherwise, simply return the argument character }
  else
    Result := c;
  end;
end;

function THTMLDocGenerator.CreateLink(Item: PItem): string;
var
  i: Integer;
begin
  Result := '';
  if (not Assigned(Item)) then Exit;
  if Assigned(Item^.MyUnit) then
    begin
      if Assigned(Item^.MyObject) then
        begin
          { it's a method, a field or a property - only those have MyObject initialized }
          // Result := Item^.MyObject^.Name + GetFileExtension + '#' + AnchorToString(Item^.AnchorNumber);
          Result := Item^.MyObject^.FullLink + '#' + AnchorToString(Item^.AnchorNumber);
        end
      else
        begin
          if (TypeOf(TCIO) = TypeOf(Item^)) then
            begin
              { it's an object / a class }
              // Result := Item^.Name + GetFileExtension;
              Result := Item^.Name + GetFileExtension;
              i := 0;
              while ExistsFullPath(Result) do
                begin
                  Inc(i);
                  Result := Item^.Name + IntToStr(i) + GetFileExtension;
                end;
            end
          else
            begin
              { it's a constant, a variable, a type or a function / procedure }
              Result := Item^.MyUnit^.FullLink + '#' + AnchorToString(Item^.AnchorNumber);
            end;
        end;
    end
  else
    begin
      { it's a unit - only units don't have a MyUnit pointer }
      Result := Item^.Name + GetFileExtension;
    end;
end;

function THTMLDocGenerator.CreateReferencedLink(ItemName, Link: string): string;
begin
  Result := CodeString('<A href="' + Link + '">' + ItemName + '</A>');
end;

function THTMLDocGenerator.ExistsFullPath(s: string): Boolean;
var
  i, j: LongInt;
  CO: PCIO;
  U: PUnit;
begin
  Result := False;

  if (not Assigned(Units)) or (Units^.Count < 1) then
    Exit;

  for i := 0 to Units^.Count - 1 do
    begin
      U := Units^.At(i);
      Result := CompareText(U^.FullLink, s) = 0;
      if Result then Exit;

      if Assigned(U^.CIO) and (U^.CIO^.Count > 0) then
        begin
          for j := 0 to U^.CIO^.Count - 1 do
            begin
              CO := U^.CIO^.At(j);
              Result := CompareText(CO^.FullLink, s) = 0;
              if Result then Exit;
            end;
        end;
    end;
end;

function THTMLDocGenerator.GetFileExtension: string;
begin
  { '.html' makes DOS version unhappy - welcome to the past ;-) }
  GetFileExtension :=
    {$IFDEF OS_DOS}
  '.htm';
  {$ELSE}
  '.htm';
  {$ENDIF}
end;

{
function THTMLDocGenerator.GetLink(ItemReference, ItemName: String): String;
begin
  GetLink := '<A href="' + ItemReference + '">' + ItemName + '</A>';
end;
}

procedure THTMLDocGenerator.WriteAppInfo;
var
  t: TTime;
begin
  { check if user does not want a link to the pasdoc homepage }
  if NoGeneratorInfo then Exit;
  { get current date/time }
  t.Init;
  { write a horizontal line, pasdoc version and a link to the pasdoc homepage }
  Stream^.WriteLine('<HR noshade size=1><EM>Generated by <A href="' + AppInfo.APP_HOMEPAGE +
    '">' + AppInfo.APP_NAME_AND_VERSION + '</A> on ' +
    t.GetDefaultString + '</EM>');
end;

procedure THTMLDocGenerator.WriteAuthors(HL: Byte; Authors: PStringCollection);
var
  i: LongInt;
  p: PString;
  s1: string;
  s2: string;
  EmailAddress: string;
begin
  if (not Assigned(Authors)) or (Authors^.Count < 1) then Exit;
  if (Authors^.Count = 1) then
    WriteHeading(HL, GetString(TEXT_AUTHOR))
  else
    WriteHeading(HL, GetString(TEXT_AUTHORS));
  for i := 0 to Authors^.Count - 1 do
    begin
      p := Authors^.At(i);
      WriteStartOfParagraph;
      if Assigned(p) then
        begin
          if ExtractEmailAddress(p^, s1, s2, EmailAddress) then
            begin
              WriteString(s1);
              Stream^.WriteString('<A href="mailto:' + EmailAddress + '">');
              WriteString(EmailAddress);
              Stream^.WriteString('</A>');
              WriteString(s2);
            end
          else
            begin
              WriteString(p^);
            end;
        end;
      WriteEndOfParagraph;
    end;
end;

procedure THTMLDocGenerator.WriteCIO(HL: Byte; var CIO: TCIO);
var
  i: LongInt;
  p: PString;
  s: string;
  Item: PCIO;
begin
  case CIO.MyType of
    CIO_CLASS: s := GetString(TEXT_CLASS);
    CIO_INTERFACE: s := GetString(TEXT_INTERFACE);
    CIO_OBJECT: s := GetString(TEXT_OBJECT);
  else
    s := '';
  end;

  s := s + ' ' + CIO.Name;

  WriteStartOfDocument(CIO.MyUnit^.Name + ': ' + s);
  WriteDocumentHeadline;

  WriteStartOfAnchor(AnchorToString(CIO.AnchorNumber));
  WriteHeading(HL, s);
  WriteEndOfAnchor;
  { write unit link }
  if Assigned(CIO.MyUnit) then
    begin
      WriteHeading(HL + 1, GetString(TEXT_UNIT));
      Stream^.WriteString('<A href="' + CIO.MyUnit^.FullLink + '">' +
        CIO.MyUnit^.Name + '</A><BR>');
    end;

  { write declaration link }
  WriteHeading(HL + 1, GetString(TEXT_DECLARATION));
  Stream^.WriteString('<P>');
  WriteStartOfCode;
  Stream^.WriteString('type ' + CIO.Name + ' = ');
  case CIO.MyType of
    CIO_CLASS: Stream^.WriteString('class');
    CIO_INTERFACE: Stream^.WriteString('interface');
  else
    Stream^.WriteString('object');
  end;
  if Assigned(CIO.Ancestors) and (CIO.Ancestors^.Count > 0) then
    begin
      Stream^.WriteString('(');
      for i := 0 to CIO.Ancestors^.Count - 1 do
        begin
          p := PString(CIO.Ancestors^.At(i));
          // RJ s := SearchLink(p^, nil);
          s := SearchLink(p^, @CIO);
          Stream^.WriteString(s);
          if (i <> CIO.Ancestors^.Count - 1) then Stream^.WriteString(', ');
        end;
      Stream^.WriteString(')');
    end;
  WriteEndOfCode;
  Stream^.WriteString('</P>');

  { Write Description }
  WriteDescription(HL + 1, GetString(TEXT_DESCRIPTION), CIO);

  { Write Hierarchy }
  if Assigned(CIO.Ancestors) and (CIO.Ancestors^.Count > 0) then
    begin
      p := PString(CIO.Ancestors^.At(0));
      Item := PCIO(SearchItem(p^, @CIO));
      if Assigned(Item) then
        begin
          WriteHeading(HL + 1, GetString(TEXT_HIERARCHY));
          repeat
            s := CreateReferencedLink(Item^.Name, Item^.FullLink);
            Stream^.WriteString(s);

            if Assigned(Item^.Ancestors) and (Item^.Ancestors^.Count > 0) then
              begin
                p := PString(Item^.Ancestors^.At(0));
                Item := PCIO(SearchItem(p^, Item));
                if Assigned(Item) then Stream^.WriteString('&nbsp;| ');
              end
            else
              Break;
          until Item = nil;
        end;
    end;

  WriteItems(HL + 1, GetString(TEXT_FIELDS), CIO.Fields);
  WriteFuncsProcs(HL + 1, True, CIO.Methods);
  if (CIO.MyType <> CIO_OBJECT) then
    WriteProperties(HL + 1, CIO.Properties);
  WriteAuthors(HL + 1, CIO.Authors);
  WriteDates(HL + 1, CIO.Created, CIO.LastMod);
  WriteFooter;
  WriteAppInfo;
  WriteEndOfDocument;
end;

procedure THTMLDocGenerator.WriteCIOs(HL: Byte; c: PItemCollection);
var
  i: LongInt;
  p: PCIO;
begin
  if (not Assigned(c)) or (c^.Count < 1) then Exit;
  for i := 0 to c^.Count - 1 do
    begin
      p := c^.At(i);
      // if (not CreateStream(p^.Name))
      if (not CreateStream(p^.OutputFileName)) then
        begin
          PrintLn(1, 'Error: Could not create class/interface/object doc file.');
          Continue;
        end;
      PrintLn(3, 'Creating class/interface/object file "' + Stream^.GetName + '"...');
      WriteCIO(HL, p^);
    end;
  CloseStream;
end;

procedure THTMLDocGenerator.WriteCIOSummary(HL: Byte; c: PItemCollection);
var
  i: LongInt;
  p: PCIO;
begin
  WriteHeading(HL, GetString(TEXT_CIO));
  if (not Assigned(c)) or (c^.Count < 1) then
    begin
      WriteStartOfParagraph;
      WriteString(GetString(TEXT_NONE));
      WriteEndOfParagraph;
      Exit;
    end;
  WriteStartOfTable2Columns(GetString(TEXT_NAME), GetString(TEXT_DESCRIPTION));
  for i := 0 to c^.Count - 1 do
    begin
      p := c^.At(i);
      WriteStartOfTableRow;
      { name of class/interface/object and unit }
      WriteStartOfTableCell;
      WriteString(GetCIOTypeName(p^.MyType));
      Stream^.WriteString('&nbsp;');
      WriteStartOfLink(p^.FullLink);
      WriteString(p^.Name);
      WriteEndOfLink;
      WriteEndOfTableCell;
      { description of class/interface/object }
      WriteStartOfTableCell;
      if Assigned(p^.Description) and (p^.Description^.Content > 0) then
        WriteText(p^.Description)
      else
        Stream^.WriteString('&nbsp;');
      WriteEndOfTableCell;
      WriteEndOfTableRow;
    end;
  WriteEndOfTable;
end;

procedure THTMLDocGenerator.WriteCodeWithLinks(p: PItem; s: string);
var
  FoundColon: Boolean;
  FoundItem: PItem;
  i: LongInt;
  l: string;
  n: Integer;
  s1: string;
  s2: string;
  S3: string;
begin
  WriteStartOfCode;
  i := 1;
  FoundColon := False;
  while (i <= Length(s)) do
    begin
      if (s[i] in ['_', 'A'..'Z', 'a'..'z']) then
        begin
          { assemble item }
          l := '';
          repeat
            l := l + s[i];
            Inc(i);
          until (i > Length(s)) or
            (not (s[i] in ['.', '_', '0'..'9', 'A'..'Z', 'a'..'z']));
          { search for item of name  L }
          if FoundColon and (SplitLink(l, s1, s2, S3, n)) then
            begin
              FoundItem := p^.FindName(s1, s2, S3, n);
              if (not Assigned(FoundItem)) then
                begin
                  FoundItem := FindGlobal(s1, s2, S3, n);
                end;
              if (FoundItem = p) then FoundItem := nil;
            end
          else
            FoundItem := nil;
          if (Assigned(FoundItem)) then
            begin
              // WriteStartOfLink(CreateLink(FoundItem));
              WriteStartOfLink(FoundItem^.FullLink);
              Stream^.WriteString(l);
              WriteEndOfLink;
            end
          else
            WriteString(l);
        end
      else
        begin
          if (s[i] = ':') then
            FoundColon := True
          else if (s[i] = ';') then
            FoundColon := False;
          WriteChar(s[i]);
          Inc(i);
        end;
    end;
  WriteEndOfCode;
end;

procedure THTMLDocGenerator.WriteComment(const s: string);
begin
  Stream^.WriteString('<!--');
  WriteString(s);
  Stream^.WriteLine('-->');
end;

procedure THTMLDocGenerator.WriteDates(HL: Byte; Created, LastMod: PString);
begin
  if Assigned(Created) then
    begin
      WriteHeading(HL, GetString(TEXT_CREATED));
      WriteStartOfParagraph;
      WriteString(Created^);
      WriteEndOfParagraph;
    end;
  if Assigned(LastMod) then
    begin
      WriteHeading(HL, GetString(TEXT_LASTMOD));
      WriteStartOfParagraph;
      WriteLine(LastMod^);
      WriteEndOfParagraph;
    end;
end;

procedure THTMLDocGenerator.WriteDocumentation;
begin
  WriteUnits(1);
  CloseStream;
  WriteOverviewFiles;
end;

procedure THTMLDocGenerator.WriteDocumentHeadline;
var
  i: Byte;
begin
  Stream^.WriteLine('<TABLE cellspacing=' + HTML_TABLE_CELLSPACING + ' cellpadding=' + HTML_TABLE_CELLPADDING + ' width=100%>');
  Stream^.WriteLine('<TR bgcolor="' + HTML_HEADER_BACKGROUND_COLOR + '">');
  for i := 0 to NUM_OVERVIEW_FILES - 1 do
    begin
      Stream^.WriteString('<TD>' +
        '<A href="' + OverviewFilenames[i] + GetFileExtension + '"><CENTER>');
      WriteString(GetString(TEXT_OVERVIEW_UNITS + i));
      Stream^.WriteLine('</CENTER></A></TD>');
    end;
  Stream^.WriteLine('</TR>');
  Stream^.WriteLine('</TABLE>');
end;

procedure THTMLDocGenerator.WriteEmptyCell;
begin
  Stream^.WriteString('&nbsp;');
end;

procedure THTMLDocGenerator.WriteEndOfDocument;
begin
  Stream^.WriteLine('</BODY>');
  Stream^.WriteLine('</HTML>');
end;

procedure THTMLDocGenerator.WriteEndOfAnchor;
begin
  Stream^.WriteString('</A>');
end;

procedure THTMLDocGenerator.WriteEndOfCode;
begin
  Stream^.WriteString('</CODE>');
end;

procedure THTMLDocGenerator.WriteEndOfLink;
begin
  Stream^.WriteString('</A>');
end;

procedure THTMLDocGenerator.WriteEndOfParagraph;
begin
  Stream^.WriteLine('</P>');
end;

procedure THTMLDocGenerator.WriteEndOfTableCell;
begin
  Stream^.WriteLine('</TD>');
end;

procedure THTMLDocGenerator.WriteEndOfTable;
begin
  Stream^.WriteLine('</TABLE>');
end;

procedure THTMLDocGenerator.WriteEndOfTableRow;
begin
  Stream^.WriteLine('</TR>');
end;

procedure THTMLDocGenerator.WriteFooter;
begin
  if HtmlFooter <> '' then
    Stream^.WriteBytes(Length(HtmlFooter), Pointer(HtmlFooter)^);
end;

procedure THTMLDocGenerator.WriteFuncsProcs(HL: Byte; Methods: Boolean;
  FuncsProcs: PMethodCollection);
var
  i: LongInt;
  j: LongInt;
  p: PMethod;
  TheLink: string;
begin
  if Methods then
    WriteHeading(HL, GetString(TEXT_METHODS))
  else
    WriteHeading(HL, GetString(TEXT_FUNCTIONS_AND_PROCEDURES));

  if (not Assigned(FuncsProcs)) or (FuncsProcs^.Count < 1) then
    begin
      WriteStartOfParagraph;
      WriteString(GetString(TEXT_NONE));
      WriteEndOfParagraph;
      Exit;
    end;

  for i := 0 to 1 do
    begin
      if (i = 0) then
        begin
          WriteHeading(HL + 1, GetString(TEXT_OVERVIEW));
          WriteStartOfTable1Column('');
        end
      else
        WriteHeading(HL + 1, GetString(TEXT_DESCRIPTION));

      for j := 0 to FuncsProcs^.Count - 1 do
        begin
          p := FuncsProcs^.At(j);
          if (i = 0) then
            begin
              WriteStartOfTableRow;

              if Methods then
                case p^.State of
                  STATE_PROTECTED:
                    begin
                      WriteStartOfTableCell;
                      Stream^.WriteString('<IMG src="protected.gif"> ');
                      WriteEndOfTableCell;
                    end;
                  STATE_PUBLIC:
                    begin
                      WriteStartOfTableCell;
                      Stream^.WriteString('<IMG src="public.gif"> ');
                      WriteEndOfTableCell;
                    end;
                  STATE_PUBLISHED:
                    begin
                      WriteStartOfTableCell;
                      Stream^.WriteString('<IMG src="published.gif"> ');
                      WriteEndOfTableCell;
                    end;
                end;

              WriteStartOfTableCell;

              TheLink := p^.FullLink;
              if Assigned(p^.MyUnit) then
                if CompareText(p^.MyUnit^.FullLink, Copy(TheLink, 1, Length(p^.MyUnit^.FullLink))) = 0 then
                  Delete(TheLink, 1, Length(p^.MyUnit^.FullLink));

              WriteStartOfLink(TheLink);
              // Stream^.WriteString('<P>');
              WriteStartOfCode;
              // WriteString (p^.Name); // The method's name only
              WriteString(p^.FullDeclaration); // The method's full declaration
              WriteEndOfCode;
              // Stream^.WriteString('</P>');
              WriteEndOfLink;
              WriteEndOfTableCell;
              WriteEndOfTableRow;
            end
          else
            begin
              WriteStartOfTable1Column('');
              WriteStartOfTableRow;
              WriteStartOfAnchor(AnchorToString(p^.AnchorNumber));
              WriteStartOfTableCell;
              WriteCodeWithLinks(p, p^.FullDeclaration);
              {Stream^.WriteString('<code>');
              WriteString(P^.FullDeclaration);
              Stream^.WriteString('</code>');}
              WriteEndOfTableCell;
              WriteEndOfAnchor;
              WriteEndOfTableRow;
              WriteEndOfTable;
              // RJ WriteEndOfParagraph;

              WriteStartOfParagraph;
              WriteItemDetailedDescription(p);
              WriteEndOfParagraph;
            end;
        end;
      if (i = 0) then WriteEndOfTable;
    end;
end;

procedure THTMLDocGenerator.WriteHeading(i: Byte; const s: string);
var
  c: Char;
begin
  if (i < 1) then i := 1;
  if (i > 6) then
    begin
      PrintLn(2, 'Warning: HTML generator cannot write headlines of level 7 or greater; will use 6 instead!');
      i := 6;
    end;
  c := Chr(i + 48);
  Stream^.WriteString('<H' + c + '>');
  WriteString(s);
  Stream^.WriteLine('</H' + c + '>');
end;

procedure THTMLDocGenerator.WriteItemDescription(Item: PItem);
begin
  if (not Assigned(Item)) then Exit;

  if Assigned(Item^.Description) and (Item^.Description^.Content > 0) then
    WriteText(Item^.Description)
  else if Assigned(Item^.DetailedDescription) and (Item^.DetailedDescription^.Content > 0) then
    WriteText(Item^.DetailedDescription)
  else
    Stream^.WriteString('&nbsp;');
end;

procedure THTMLDocGenerator.WriteItemDetailedDescription(Item: PItem);
begin
  if (not Assigned(Item)) then Exit;

  if Assigned(Item^.DetailedDescription) and (Item^.DetailedDescription^.Content > 0) then
    WriteText(Item^.DetailedDescription)
  else if Assigned(Item^.Description) and (Item^.Description^.Content > 0) then
    WriteText(Item^.Description)
  else
    Stream^.WriteString('&nbsp;');
end;

procedure THTMLDocGenerator.WriteItems(Order: Byte; Heading: string; i: PItemCollection);
var
  j: LongInt;
  Item: PItem;
begin
  WriteHeading(Order, Heading);
  if (not Assigned(i)) or (i^.Count < 1) then
    begin
      Stream^.WriteString('<P>');
      WriteString(GetString(TEXT_NONE));
      Stream^.WriteString('</P>');
      Exit;
    end;
  Stream^.WriteString('<TABLE cellspacing=' + HTML_TABLE_CELLSPACING + ' cellpadding=' + HTML_TABLE_CELLPADDING + ' width=100%>');
  Stream^.WriteString('<TR bgcolor="#' + HTML_HEADER_BACKGROUND_COLOR + '">');
  Stream^.WriteLine('<TH>' + GetString(TEXT_NAME) + '</TH><TH>' + GetString(TEXT_DESCRIPTION) + '</TH></TR>');
  j := 0;
  repeat
    Item := i^.At(j);
    WriteStartOfTableRow;
    WriteStartOfAnchor(AnchorToString(Item^.AnchorNumber));
    WriteStartOfTableCell;
    WriteString(Item^.Name);
    WriteEndOfTableCell;
    WriteStartOfTableCell;
    WriteItemDetailedDescription(Item);
    WriteEndOfTableCell;
    WriteEndOfAnchor;
    WriteEndOfTableRow;
    Inc(j);
  until (j = i^.Count);
  Stream^.WriteString('</TABLE>');
end;

procedure THTMLDocGenerator.WriteOverviewFiles;
var
  c1: PItemCollection;
  c2: PItemCollection;
  c3: PItemCollection;
  i: LongInt;
  Item: PItem;
  j: LongInt;
  PU: PUnit;
begin
  WriteUnitOverviewFile;
  if (not Assigned(Units)) or (Units^.Count < 1) then Exit;
  c3 := New(PItemCollection, Init(16, 16));
  for i := 1 to 5 do
    begin
      // if (not CreateStream(OverviewFilenames[i]))
      if (not CreateStream(DestDir + OverviewFilenames[i] + GetFileExtension)) then
        begin
          PrintLn(1, 'Error: Could not create output file "' +
            OverviewFilenames[i] + '".');
          Exit;
        end;
      PrintLn(3, 'Writing overview file ' + Stream^.GetName + '...');
      WriteStartOfDocument(GetString(TEXT_HEADLINE_CIO + i - 1));
      WriteDocumentHeadline;
      WriteHeading(1, GetString(TEXT_HEADLINE_CIO + i - 1));
      c1 := New(PItemCollection, Init(16, 16));

      // Allow for Duplicates in Classes
      if i = 1 then
        c1^.Duplicates := True
      else
        c1^.Duplicates := False;

      for j := 0 to Units^.Count - 1 do
        begin
          PU := Units^.At(j);
          case i of
            1: c2 := PU^.CIO;
            2: c2 := PU^.Types;
            3: c2 := PU^.Variables;
            4: c2 := PU^.Constants;
            5: c2 := PU^.FuncsProcs;
          end;
          c1^.InsertItems(c2);
        end;
      if Assigned(c1) and (c1^.Count > 0) then
        begin
          WriteStartOfTable3Columns(GetString(TEXT_NAME), GetString(TEXT_UNIT), GetString(TEXT_DESCRIPTION));
          for j := 0 to c1^.Count - 1 do
            begin
              Item := c1^.At(j);
              WriteStartOfTableRow;
              WriteStartOfTableCell;
              WriteStartOfLink(Item^.FullLink);
              WriteString(Item^.Name);
              WriteEndOfLink;
              WriteEndOfTableCell;
              WriteStartOfTableCell;
              WriteStartOfLink(Item^.MyUnit^.FullLink);
              WriteString(Item^.MyUnit^.Name);
              WriteEndOfLink;
              WriteEndOfTableCell;
              WriteStartOfTableCell;
              WriteItemDescription(Item);
              WriteEndOfTableCell;
              WriteEndOfTableRow;
            end;
          WriteEndOfTable;
        end
      else
        begin
          WriteStartOfParagraph;
          WriteString(GetString(TEXT_NONE));
          WriteEndOfParagraph;
        end;
      c3^.InsertItems(c1);
      Dispose(c1, Done);
      WriteFooter;
      WriteAppInfo;
      WriteEndOfDocument;
      CloseStream;
    end;
  // if (not CreateStream(OverviewFilenames[6]))
  if (not CreateStream(DestDir + OverviewFilenames[6] + GetFileExtension)) then
    begin
      PrintLn(1, 'Error: Could not create overview output file "' +
        OverviewFilenames[6] + '".');
      Exit;
    end;
  PrintLn(3, 'Writing overview file ' + Stream^.GetName + '...');
  WriteStartOfDocument(GetString(TEXT_HEADLINE_IDENTIFIERS));
  WriteDocumentHeadline;
  WriteHeading(1, GetString(TEXT_HEADLINE_IDENTIFIERS));
  WriteStartOfTable3Columns(GetString(TEXT_NAME),
    GetString(TEXT_UNIT), GetString(TEXT_DESCRIPTION));
  for i := 0 to c3^.Count - 1 do
    begin
      Item := c3^.At(i);
      WriteStartOfTableRow;

      WriteStartOfTableCell;
      WriteStartOfLink(Item^.FullLink);
      WriteString(Item^.Name);
      WriteEndOfLink;
      WriteEndOfTableCell;

      WriteStartOfTableCell;
      WriteStartOfLink(Item^.MyUnit^.FullLink);
      WriteString(Item^.MyUnit^.Name);
      WriteEndOfLink;
      WriteEndOfTableCell;

      WriteStartOfTableCell;
      WriteItemDescription(Item);
      WriteEndOfTableCell;

      WriteEndOfTableRow;
    end;
  WriteEndOfTable;
  WriteFooter;
  WriteAppInfo;
  WriteEndOfDocument;
  CloseStream;
end;

procedure THTMLDocGenerator.WriteParagraph(HL: Byte; s: string; t: PText);
begin
  if (not Assigned(t)) or (t^.Content < 1) then Exit;
  WriteHeading(HL, s);
  Stream^.WriteLine('<P>');
  Stream^.WriteBytes(t^.Content, t^.Data^);
  Stream^.WriteLine('</P>');
end;

procedure THTMLDocGenerator.WriteProperties(HL: Byte; p: PPropertyCollection);
var
  i: LongInt;
  j: LongInt;
  Prop: PProperty;
begin
  WriteHeading(HL, GetString(TEXT_PROPERTIES));
  if (not Assigned(p)) or (p^.Count < 1) then
    begin
      Stream^.WriteLine('<P>');
      WriteString(GetString(TEXT_NONE));
      Stream^.WriteLine('</P>');
      Exit;
    end;
  for i := 0 to 1 do
    begin
      if (i = 0) then
        WriteHeading(HL + 1, GetString(TEXT_OVERVIEW))
      else
        WriteHeading(HL + 1, GetString(TEXT_DESCRIPTION));
      for j := 0 to p^.Count - 1 do
        begin
          Prop := p^.At(j);
          WriteStartOfTable1Column('');
          WriteStartOfTableRow;
          WriteStartOfTableCell;
          if (i = 0) then
            Stream^.WriteString(CreateReferencedLink(Prop^.FullDeclaration, Prop^.FullLink))
          else
            begin
              WriteStartOfAnchor(AnchorToString(Prop^.AnchorNumber));
              {WriteString(Prop^.FullDeclaration);}
              WriteCodeWithLinks(Prop, Prop^.FullDeclaration);
              {Stream^.WriteString(S);}
              WriteEndOfAnchor;
            end;
          WriteEndOfTableCell;
          WriteEndOfTableRow;
          WriteEndOfTable;
          if (i = 1) then
            begin
              WriteStartOfParagraph;
              WriteItemDetailedDescription(Prop);
              WriteEndOfParagraph;
            end;
        end;
    end;
end;

procedure THTMLDocGenerator.WriteStartOfAnchor(const Name: string);
begin
  Stream^.WriteString('<A name="' + Name + '">');
end;

procedure THTMLDocGenerator.WriteStartOfCode;
begin
  Stream^.WriteString('<CODE>');
end;

procedure THTMLDocGenerator.WriteStartOfDocument(Name: string);
begin
  Stream^.WriteLine('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN" "http://www.w3.org/TR/REC-html40/loose.dtd">');
  Stream^.WriteLine('<HTML>');
  Stream^.WriteLine('<HEAD>');
  Stream^.WriteString('<TITLE>');
  WriteString(Name);
  Stream^.WriteLine('</TITLE>');
  Stream^.WriteLine('</HEAD');
  Stream^.WriteLine('<BODY bgcolor="#ffffff" text="#000000" link="#0000ff" vlink="#800080" alink="#FF0000">');

  if HtmlHeader <> '' then
    Stream^.WriteBytes(Length(HtmlHeader), Pointer(HtmlHeader)^);
end;

procedure THTMLDocGenerator.WriteStartOfLink(const Name: string);
begin
  Stream^.WriteString('<A href="' + Name + '">');
end;

procedure THTMLDocGenerator.WriteStartOfParagraph;
begin
  Stream^.WriteString('<P>');
end;

procedure THTMLDocGenerator.WriteStartOfTable1Column(t: string);
begin
  Stream^.WriteLine('<TABLE cellspacing=' + HTML_TABLE_CELLSPACING + ' cellpadding=' + HTML_TABLE_CELLPADDING + ' width=100%>');
end;

procedure THTMLDocGenerator.WriteStartOfTable2Columns(t1, t2: string);
begin
  Stream^.WriteLine('<TABLE cellspacing=' + HTML_TABLE_CELLSPACING + ' cellpadding=' + HTML_TABLE_CELLPADDING + ' width=100%>');
  Stream^.WriteString('<TR bgcolor="#' + HTML_HEADER_BACKGROUND_COLOR + '"><TH>');
  WriteString(t1);
  Stream^.WriteString('</TH><TH>');
  WriteString(t2);
  Stream^.WriteLine('</TH></TR>');
end;

procedure THTMLDocGenerator.WriteStartOfTable3Columns(t1, t2, T3: string);
begin
  Stream^.WriteLine('<TABLE cellspacing=' + HTML_TABLE_CELLSPACING + ' cellpadding=' + HTML_TABLE_CELLPADDING + ' width=100%>');
  Stream^.WriteString('<TR bgcolor="#' + HTML_HEADER_BACKGROUND_COLOR + '"><TH>');
  WriteString(t1);
  Stream^.WriteString('</TH><TH>');
  WriteString(t2);
  Stream^.WriteString('</TH><TH>');
  WriteString(T3);
  Stream^.WriteLine('</TH></TR> ');
end;

procedure THTMLDocGenerator.WriteStartOfTableCell;
begin
  Stream^.WriteString('<TD>');
end;

procedure THTMLDocGenerator.WriteStartOfTableRow;
begin
  Stream^.WriteString('<TR bgcolor=#' + HTML_ROW_BACKGROUND_COLOR + ' valign=top>');
end;

procedure THTMLDocGenerator.WriteUnit(HL: Byte; U: PUnit);
var
  d: PText;
begin
  if (not Assigned(U)) then
    begin
      PrintLn(1, 'Error in THTMLDocGenerator.WriteUnit: Unit variable has not been initialized.');
      Exit;
    end;
  // if (not CreateStream(u^.Name)) then
  if (not CreateStream(U^.OutputFileName)) then
    begin
      PrintLn(1, 'Error: Could not create HTML unit doc file for unit ' + U^.Name);
      Exit;
    end;
  PrintLn(3, 'Writing documentation for unit "' + U^.Name + '"...');
  WriteStartOfDocument(U^.Name);
  WriteDocumentHeadline;
  WriteHeading(HL, GetString(TEXT_UNIT) + ' ' + U^.Name);

  if (Assigned(U^.DetailedDescription)) then
    d := U^.DetailedDescription
  else
    d := U^.Description;

  WriteUnitDescription(HL + 1, d);
  WriteCIOSummary(HL + 1, U^.CIO);
  WriteFuncsProcs(HL + 1, False, U^.FuncsProcs);
  WriteTypes(HL + 1, U^.Types);
  WriteConstants(HL + 1, U^.Constants);
  WriteVariables(HL + 1, U^.Variables);
  WriteAuthors(HL + 1, U^.Authors);
  WriteDates(HL + 1, U^.Created, U^.LastMod);
  WriteFooter;
  WriteAppInfo;
  WriteEndOfDocument;
  CloseStream;
  WriteCIOs(HL, U^.CIO);
end;

procedure THTMLDocGenerator.WriteUnitDescription(HL: Byte; d: PText);
begin
  WriteParagraph(HL, GetString(TEXT_DESCRIPTION), d);
end;

procedure THTMLDocGenerator.WriteUnitOverviewFile;
var
  c: PItemCollection;
  Item: PItem;
  j: LongInt;
begin
  c := Units;
  if (not CreateStream(DestDir + OverviewFilenames[0] + GetFileExtension)) then
    begin
      PrintLn(1, 'Error: Could not create overview output file "' + OverviewFilenames[0] + '".');
      Exit;
    end;
  PrintLn(3, 'Writing unit overview file "' + Stream^.GetName + '"...');
  WriteStartOfDocument(GetString(TEXT_HEADLINE_UNITS));
  WriteDocumentHeadline;
  WriteHeading(1, GetString(TEXT_HEADLINE_UNITS));
  if Assigned(c) and (c^.Count > 0) then
    begin
      WriteStartOfTable2Columns(GetString(TEXT_NAME), GetString(TEXT_DESCRIPTION));
      for j := 0 to c^.Count - 1 do
        begin
          Item := c^.At(j);
          WriteStartOfTableRow;
          WriteStartOfTableCell;
          WriteStartOfLink(Item^.FullLink);
          WriteString(Item^.Name);
          WriteEndOfLink;
          WriteEndOfTableCell;
          WriteStartOfTableCell;
          WriteItemDescription(Item);
          WriteEndOfTableCell;
          WriteEndOfTableRow;
        end;
      WriteEndOfTable;
    end;
  WriteFooter;
  WriteAppInfo;
  WriteEndOfDocument;
  CloseStream;
end;

begin
end.

