{
@abstract(contains the application object TPasdoc)
@author(Erwin Scheuch-Heilig (ScheuchHeilig@t-online.de))
@author(Marco Schmidt (marcoschmidt@geocities.com))
@author(Michael van Canneyt (michael@tfdec1.fys.kuleuven.ac.be))
@created(24 Sep 1999)
@lastmod(20 Apr 2000)
}
unit Main;

{$I platform.inc}

interface

{ Wildcard expansion is currently supported for any DOS compiler (FPC/DOS,
  BP) and Delphi.
  Linux doesn't need this, as the shell does the expansion. }

{$IFDEF PPC_DELPHI}
{ tell Delphi that this is a console application }
{$APPTYPE CONSOLE}
{ file search using SysUtils }
{$DEFINE SUPPORT_WILDCARD_EXPANSION}
{$ENDIF}

{$IFDEF OS_DOS}
{ file search using DOS unit }
{$DEFINE SUPPORT_WILDCARD_EXPANSION}
{$DEFINE USE_DOS}
{$ENDIF}

{$IFDEF PPC_FPC}
{ file search using DOS unit }
{$DEFINE SUPPORT_WILDCARD_EXPANSION}
{$DEFINE USE_DOS}
{$ENDIF}

uses
  {$IFDEF USE_DOS}
  DOS,
  {$ENDIF}
  {$IFDEF PPC_DELPHI}
  SysUtils,
  {$ENDIF}
  AppInfo,
  Arrays,
  Chars,
  FileStre,
  GenDoc,
  Html,
  Items,
  Msg,
  Numbers,
  Objects,
  Parsing,
  Streams,
  Tex,
  Texts,
  Time;

type
  { all supported output formats }
  TOutputFormat = (OUTPUT_NONE, OUTPUT_LATEX, OUTPUT_HTML);
  { pointer to @link(TPasDoc) }
  PPasDoc = ^TPasDoc;
  { The main object in the pasdoc application; first scans parameters, then
    parses files.
    All parsed units are then given to documentation generator,
    which creates one or more documentation output files. }
  TPasDoc = object(TObject)
    { all description file names }
    DescriptionFiles: PStringCollection;
    { directory where documentation will be written to }
    DestDir: string;
    { all file names which have been collected from the program
      parameters }
    FileNames: PStringCollection;
    HtmlHeader: AnsiString;
    HtmlFooter: AnsiString;
    { collection of @link(PString) values of all directories where pasdoc is
      supposed to search for include files }
    IncludeFilePaths: PCollection;
    { defines if the private parts (fields and methods) of objects are to be
      included in the documentation (default is false) }
    IncludePrivate: Boolean;
    KeepLineFeeds: Boolean;
    { language used to write documentation }
    Language: TLanguage;
    { stores if the Language field was ever initialized from the parameters }
    LanguageInitialized: Boolean;
    { by default, all HTML documents contain a link on the bottom stating
      pasdoc version, time this page was generated and a link to the pasdoc
      homepage; set NoGeneratorInfo to true to prevent this }
    NoGeneratorInfo: Boolean;
    { if true: leave out statements at the beginning and end of TeX document
      to enable users to simply \include the pasdoc output; default is false }
    NoTexHeadersAndFooters: Boolean;
    { output format to be used (default is HTML) }
    OutputFormat: TOutputFormat;
    { list of standard compiler directives }
    StdDirectives: PStringCollection;
    { all PUnit objects which have been created from the list of file names }
    Units: PItemCollection;
    { Creates object and sets fields to default values. }
    constructor Init;
    { }
    destructor Done; virtual;
    { Adds S to the list of description files that will be parsed.
      Will compare S to strings in list and will not include it if already
      present.
      However, this will not avoid duplicates on file systems where case is
      not regarded (VFAT etc.). }
    function AddDescriptionFile(s: string): Boolean;
    { Adds parameter S to the internal list of standard directives, which
      are considered present whenever parsing of a unit starts. }
    function AddStandardDirective(s: string): Boolean;
    { Creates a @link(TUnit) object from the stream and adds it to
   @link(Units). }
    procedure HandleStream(InputStream: PInputStream);
    { Scans the program parameters, creating a list of file names and
      regarding switches. }
    procedure HandleParams;
    { Loads names of Pascal unit source code files from a text file.
      Adds all file names to @link(FileNames). }
    procedure LoadFileNames(FileName: string);
    { Loads a list of standard directives (one per row) from text file
      FileName.
      These directives are added to the standard directives list
      @link(StdDirectives). }
    procedure LoadStandardDirectives(FileName: string);
    { Calls @link(HandleStream) for each file name in @link(FileNames). }
    procedure LoadAnsiStringFromFile(FileName: string; var s: AnsiString);
    procedure ParseFiles;
    { Prints a detailed list of switches to standard output. }
    procedure PrintHelp;
    { Prints a usage clause to standard output. }
    procedure PrintUsage;
    { Searches the description of each PUnit item in the collection for an
      excluded tag.
      If one is found, the item is removed from the collection.
      If not, the fields, methods and properties collections are called
      with RemoveExcludedItems
      If the collection is empty after removal of all items, it is disposed
      of and the variable is set to nil. }
    procedure RemoveExcludedItems(var c: PItemCollection);
    { Does all the work. }
    procedure Run;
    { Searches for descr tags in the comments of all TItem objects in C. }
    procedure SearchDescrFileTags(var c: PItemCollection);
    { Sets language to NewLanguage. }
    procedure SetLanguage(const NewLanguage: TLanguage);
  end;

implementation

uses
  Windows;

const
  { the default output language (type @link(TLanguage) }
  DEFAULT_LANGUAGE: TLanguage = Chars.LANG_ENGLISH;
  { the output format to be used when none is specified on command line }
  DEFAULT_OUTPUT_FORMAT: TOutputFormat = OUTPUT_HTML;
  { names of all output formats }
  OUTPUT_FORMAT_NAMES: array[TOutputFormat] of string =
  ('', 'LaTeX', 'HTML');
  { the string used to separate paths, a backslash on Windows and a forward
    slash on Unix systems }
  {$IFDEF OS_LINUX}
  PATH_SEPARATOR = '/';
  {$ELSE}
  PATH_SEPARATOR = '\';
  {$ENDIF}

  { find the right-most character 'CharToFind' within string 'Str'
    returns 0 if character not found at all }
function RScan(CharToFind: Char; Str: string): Integer;
var
  CurrPos: Integer;
begin
  CurrPos := Length(Str);
  while (CurrPos > 0) and (Str[CurrPos] <> CharToFind) do
    Dec(CurrPos);
  RScan := CurrPos;
end;

constructor TPasDoc.Init;
begin
  inherited Init;
  { check for Unicode }
  if (SizeOf(Char) <> 1) then
    begin
      PrintLn(1, 'Error: Size of Char is larger than one byte, cannot handle this case.');
      Halt(1);
    end;
  DestDir := '';
  DescriptionFiles := nil;
  FileNames := nil;
  IncludeFilePaths := nil;
  IncludePrivate := False;
  KeepLineFeeds := False;
  Language := LANG_UNDEFINED;
  LanguageInitialized := False;
  OutputFormat := DEFAULT_OUTPUT_FORMAT;
  NoGeneratorInfo := False;
  NoTexHeadersAndFooters := False;
  StdDirectives := nil;
  Units := nil;
end;

destructor TPasDoc.Done;
begin
  if Assigned(DescriptionFiles) then Dispose(DescriptionFiles, Done);
  if Assigned(FileNames) then Dispose(FileNames, Done);
  if Assigned(Units) then Dispose(Units, Done);
  inherited Done;
end;

function TPasDoc.AddDescriptionFile(s: string): Boolean;
var
  PS: PString;
begin
  Result := False;
  if (not Assigned(DescriptionFiles)) then
    begin
      DescriptionFiles := New(PStringCollection, Init(16, 16));
      if (not Assigned(DescriptionFiles)) then
        begin
          PrintLn(2, 'Errorr: not enough memory to create description file list');
          Exit;
        end;
    end;
  if (DescriptionFiles^.IndexOf(@s) <> -1) then
    begin
      Result := True;
      Exit;
    end;
  PS := NewStr(s);
  if Assigned(PS) then
    DescriptionFiles^.Insert(PS)
  else
    PrintLn(2, 'Error - could not create description file name string');
  Result := Assigned(PS);
end;

function TPasDoc.AddStandardDirective(s: string): Boolean;
var
  PS: PString;
begin
  { convert S to upper case for convenient comparison }
  StringToUpper(s, s);
  { check if S is already included - if so, leave }
  if (Assigned(StdDirectives) and (StdDirectives^.IndexOf(@s) <> -1)) then
    begin
      AddStandardDirective := True;
      Exit;
    end;
  { create StdDirectives collection if necessary }
  if (not Assigned(StdDirectives)) then
    begin
      StdDirectives := New(PStringCollection, Init(16, 16));
      if (not Assigned(StdDirectives)) then
        begin
          PrintLn(1, 'Error: Could not create list of standard directives.');
          Halt(1);
        end;
    end;
  { create PString version of S and insert it into collection }
  PS := NewStr(s);
  if (not Assigned(PS)) then
    begin
      PrintLn(1, 'Error: Could not create new string for directive list');
      Halt(1);
    end;
  StdDirectives^.Insert(PS);
  AddStandardDirective := True;
end;

procedure TPasDoc.HandleParams;
var
  ErrorCode: Integer;
  i: LongInt;
  j: LongInt;
  n: LongInt;
  p: PString;
  Path: string;
  s: string;
  VerbosityLevel: Integer;
  {$IFDEF USE_DOS}
  SR: DOS.SearchRec;
  {$ENDIF}
  {$IFDEF PPC_DELPHI}
  SearchResult: Integer;
  SR: SysUtils.TSearchRec;
  {$ENDIF}
begin
  n := ParamCount;
  if (n < 1) then
    begin
      PrintUsage;
      PrintLn(1, 'Try ''' + ParamStr(0) + ' --help'' for more information.');
      Halt(0);
    end;
  FileNames := New(PStringCollection, Init(16, 16));
  OutputFormat := OUTPUT_NONE;
  VerbosityLevel := DEFAULT_VERBOSITY_LEVEL;
  i := 1;
  repeat
    s := ParamStr(i);
    if (s[1] = '-') then
      begin
        System.Delete(s, 1, 1);
        if (s = '-help') or (s = 'H') then
          begin
            PrintUsage;
            PrintHelp;
            Halt(0);
          end
        else
          if (s = '-version') or (s = 'V') then
            begin
              PrintLn(1, APP_NAME_AND_VERSION);
              Halt(0);
            end
          else
            if (s = 'a') or (s = '-catalan') then
              SetLanguage(Chars.LANG_CATALAN)
            else
              if (s = 'b') or (s = '-brasilian') then
                SetLanguage(Chars.LANG_BRASILIAN)
              else
                if (s = 'c') or (s = '-descrfile') then
                  begin
                    if (i = n) then
                      begin
                        PrintLn(1, 'Error: No file name specified after -c switch.');
                        Halt(1);
                      end;
                    Inc(i);
                    if (not AddDescriptionFile(ParamStr(i))) then
                      begin
                        PrintLn(1, 'Error: Could not add description file.');
                        Halt(1);
                      end;
                  end
                else
                  if (s = 'd') or (s = '-directive') then
                    begin
                      if (i = n) then
                        begin
                          PrintLn(1, 'Error: No directive specified after -d switch.');
                          Halt(1);
                        end;
                      Inc(i);
                      if (not AddStandardDirective(ParamStr(i))) then
                        begin
                          PrintLn(1, 'Error: Could not add standard directive.');
                          Halt(1);
                        end;
                    end
                  else
                    if (s = 'e') or (s = '-english') then
                      SetLanguage(Chars.LANG_ENGLISH)
                    else
                      if (s = 'f') or (s = '-directivefile') then
                        begin
                          if (i = n) then
                            begin
                              PrintLn(1, 'Error: No directive file specified after -f switch.');
                              Halt(1);
                            end;
                          Inc(i);
                          LoadStandardDirectives(ParamStr(i));
                        end
                      else
                        if (s = 'g') or (s = '-german') then
                          SetLanguage(Chars.LANG_GERMAN)
                        else
                          if (s = 'h') or (s = '-html') then
                            begin
                              if (OutputFormat <> OUTPUT_NONE) and (OutputFormat <> OUTPUT_HTML) then
                                begin
                                  PrintLn(1, 'Error: Cannot handle more than one documentation output type.');
                                  Halt(1);
                                end;
                              OutputFormat := OUTPUT_HTML;
                            end
                          else
                            if CompareText(s, 'htmlheader') = 0 then
                              begin
                                if (i = n) then
                                  begin
                                    PrintLn(1, 'Error: No filename specified after -HtmlHeader switch.');
                                    Halt(1);
                                  end;
                                Inc(i);
                                LoadAnsiStringFromFile(ParamStr(i), HtmlHeader);
                              end
                            else
                              if CompareText(s, 'htmlfooter') = 0 then
                                begin
                                  if (i = n) then
                                    begin
                                      PrintLn(1, 'Error: No filename specified after -HtmlFooter switch.');
                                      Halt(1);
                                    end;
                                  Inc(i);
                                  LoadAnsiStringFromFile(ParamStr(i), HtmlFooter);
                                end
                              else
                                if (s = 'i') or (s = '-spanish') then
                                  SetLanguage(Chars.LANG_SPANISH)
                                else
                                  if (s = 'j') or (s = '-notexheaders') then
                                    NoTexHeadersAndFooters := True
                                  else
                                    if (s = 'k') or (s = '-keeplinefeeds') then
                                      KeepLineFeeds := True
                                    else
                                      if (s = 'l') or (s = '-latex') then
                                        begin
                                          if (OutputFormat <> OUTPUT_NONE) and (OutputFormat <> OUTPUT_LATEX) then
                                            begin
                                              PrintLn(1, 'Error: Cannot handle more than one documentation output type.');
                                              Halt(1);
                                            end;
                                          OutputFormat := OUTPUT_LATEX;
                                        end
                                      else
                                        if (s = 'm') or (s = '-dutch') then
                                          SetLanguage(Chars.LANG_DUTCH)
                                        else
                                          if (s = 'n') or (s = '-nohomepage') then
                                            NoGeneratorInfo := True
                                          else
                                            if (s = 'o') or (s = '-outputdirectory') then
                                              begin
                                                if (i = n) then
                                                  begin
                                                    PrintLn(1, 'Error: No directory specified after -o switch.');
                                                    Halt(1);
                                                  end;
                                                Inc(i);
                                                DestDir := ParamStr(i);
                                                if (Length(DestDir) > 0) and (DestDir[Length(DestDir)] <> PATH_SEPARATOR) then
                                                  DestDir := DestDir + PATH_SEPARATOR;
                                              end
                                            else
                                              if (s = 'p') or (s = '-includeprivate') then
                                                IncludePrivate := True
                                              else
                                                if (s = 'r') or (s = '-french') then
                                                  SetLanguage(Chars.LANG_FRENCH)
                                                else
                                                  if (s = 's') or (s = '-sourcefilenames') then
                                                    begin
                                                      if (i = n) then
                                                        begin
                                                          PrintLn(1, 'Error: No file name specified after -s switch.');
                                                          Halt(1);
                                                        end;
                                                      Inc(i);
                                                      LoadFileNames(ParamStr(i));
                                                    end
                                                  else
                                                    if (s = 'u') or (s = '-includefilepath') then
                                                      begin
                                                        if (i = n) then
                                                          begin
                                                            PrintLn(1, 'Error: No directory specified after -u switch.');
                                                            Halt(1);
                                                          end;
                                                        Inc(i);
                                                        s := ParamStr(i);
                                                        if (Length(s) > 0) and (s[Length(s)] <> PATH_SEPARATOR) then s := s + PATH_SEPARATOR;
                                                        if (not Assigned(IncludeFilePaths)) then
                                                          begin
                                                            IncludeFilePaths := New(PCollection, Init(16, 16));
                                                            if (not Assigned(IncludeFilePaths)) then
                                                              begin
                                                                PrintLn(1, 'Error: Could not create list of include file paths.');
                                                                Halt(1);
                                                              end;
                                                          end;
                                                        p := NewStr(s);
                                                        if (not Assigned(p)) then
                                                          begin
                                                            PrintLn(1, 'Error: Could not create string for include file path.');
                                                            Halt(1);
                                                          end;
                                                        IncludeFilePaths^.Insert(p);
                                                      end
                                                    else
                                                      if (s = 'v') or (s = '-verbose') then
                                                        begin
                                                          if (i = n) then
                                                            begin
                                                              PrintLn(1, 'Error: No verbosity level (integer >= 0) specified after -v switch.');
                                                              Halt(1);
                                                            end;
                                                          Inc(i);
                                                          Val(ParamStr(i), VerbosityLevel, ErrorCode);
                                                          if (ErrorCode <> 0) or (VerbosityLevel < 0) then
                                                            begin
                                                              PrintLn(1, 'Error - verbosity level value must be an integer >= 0');
                                                              Halt(1);
                                                            end;
                                                        end
                                                      else
                                                        begin
                                                          PrintLn(1, 'Error: Unknown switch ("-' + s + '")');
                                                          Halt(1);
                                                        end;
      end
    else
      begin
        { it's not a switch, so it must be a unit source code file name or a wildcard }
        { check for wildcards }
        if (System.Pos('*', s) > 0) or (System.Pos('?', s) > 0) then
          begin
            {$IFDEF SUPPORT_WILDCARD_EXPANSION}
            Path := '';
            { Path = everything before right-most path delimiter char }
            if (RScan(PATH_SEPARATOR, s) > 0) then Path := System.Copy(s, 1, RScan(PATH_SEPARATOR, s));
            {$IFDEF USE_DOS}
            DOS.FindFirst(s, 63, SR);
            while (DOS.DosError = 0) do
              begin
                if ((SR.Attr and 24) = 0) then FileNames^.Insert(NewStr(Path + SR.Name));
                DOS.FindNext(SR);
              end;
            {$ENDIF}
            {$IFDEF PPC_DELPHI}
            SearchResult := SysUtils.FindFirst(s, 63, SR);
            while (SearchResult = 0) do
              begin
                if ((SR.Attr and 24) = 0) then FileNames^.Insert(NewStr(Path + SR.Name));
                SearchResult := FindNext(SR);
              end;
            SysUtils.FindClose(SR);
            {$ENDIF}
            {$ELSE}
            PrintLn(1, 'Warning: Wildcard expansion not supported, skipping ',
              s);
            {$ENDIF}
          end
        else
          begin
            p := NewStr(s);
            j := FileNames^.IndexOf(p);
            if (j < 0) then
              FileNames^.Insert(p)
            else
              DisposeStr(p);
          end;
      end;
    Inc(i);
  until (i > n);
  SetVerbosityLevel(VerbosityLevel);
  if (FileNames^.Count < 1) then
    begin
      PrintLn(1, 'Error: No file has been specified.');
      Halt(1);
    end;
  PrintLn(2, APP_FIRST_LINE);
  { if no output format has been defined on the command line, pick
    default output format, show warning }
  if (OutputFormat = OUTPUT_NONE) then
    begin
      PrintLn(2, 'Warning: No output format specified, using default ("' +
        OUTPUT_FORMAT_NAMES[DEFAULT_OUTPUT_FORMAT] + '").');
      OutputFormat := DEFAULT_OUTPUT_FORMAT;
    end;
  if (Language = LANG_UNDEFINED) then
    begin
      SetLanguage(DEFAULT_LANGUAGE);
      PrintLn(2, 'Warning: No output language specified, using default ("' +
        LANGUAGE_NAMES[Language] + '").');
    end;
  {  if Assigned(IncludeFilePaths) and (IncludeFilePaths^.Count > 0) then
    begin
      for I := 0 to IncludeFilePaths^.Count - 1 do
      begin
        P := IncludeFilePaths^.At(I);
        if Assigned(P)
        then WriteLn(I, ' ', P^);
      end;
    end;}
end;

procedure TPasDoc.HandleStream(InputStream: PInputStream);
var
  p: PParser;
  U: PUnit;
begin
  p := New(PParser, Init(InputStream, StdDirectives, IncludeFilePaths));
  if (not Assigned(p)) then
    begin
      PrintLn(1, InputStream^.GetName + ' Error: Could not create parser.');
      Exit;
    end;
  if p^.ParseUnit(U) then
    begin
      if (not IncludePrivate) then U^.RemovePrivateItems;

      if (not Assigned(Units)) then Units := New(PItemCollection, Init(16, 16));
      Units^.Insert(U);
    end
  else
    begin
      PrintLn(1, 'Error parsing: ' + p^.ErrorMessage);
    end;
  Dispose(p, Done);
end;

procedure TPasDoc.LoadFileNames(FileName: string);
var
  nf: PFileInputStream;
  l: string;
  s: PString;
begin
  nf := New(PFileInputStream, Init(FileName));
  if (not Assigned(nf)) then
    begin
      PrintLn(2, 'Error: Could not open file "' + FileName + '" to read file names.');
      Halt(1);
    end;
  while (nf^.HasData) do
    begin
      nf^.ReadLine(l);
      s := NewStr(l);
      if (not Assigned(s)) then
        begin
          PrintLn(2, 'Error: Not enough memory to create list of file names.');
          Halt(1);
        end;
      if (FileNames^.IndexOf(@s) = -1) then
        FileNames^.Insert(s)
      else
        DisposeStr(s);
    end;
  Dispose(nf, Done);
end;

procedure TPasDoc.LoadStandardDirectives(FileName: string);
var
  DF: PFileInputStream;
  l: string;
begin
  DF := New(PFileInputStream, Init(FileName));
  if (not Assigned(DF)) then
    begin
      PrintLn(1, 'Error: Could not open file "' + FileName + '" to read standard directives.');
      Halt(1);
    end;
  while (DF^.HasData) do
    begin
      DF^.ReadLine(l);
      if (not AddStandardDirective(l)) then Break;
    end;
  Dispose(DF, Done);
end;

procedure TPasDoc.LoadAnsiStringFromFile(FileName: string; var s: AnsiString);
var
  nf: PFileInputStream;
  l: Integer;
begin
  nf := New(PFileInputStream, Init(FileName));
  if (not Assigned(nf)) then
    begin
      PrintLn(2, 'Error: Could not open file "' + FileName + '" for reading.');
      Halt(1);
    end;

  l := nf^.GetSize;
  SetLength(s, l);
  nf^.ReadRawBytes(l, Pointer(s)^);
  Dispose(nf, Done);
end;

procedure TPasDoc.ParseFiles;
var
  i: LongInt;
  p: PString;
  InputStream: PFileInputStream;
begin
  if (not Assigned(FileNames)) or (FileNames^.Count < 1) then Exit;
  i := 0;
  repeat
    p := FileNames^.At(i);
    PrintLn(3, 'Loading "' + p^ + '"...');
    InputStream := New(PFileInputStream, Init(p^));
    if (not Assigned(InputStream)) then
      begin
        PrintLn(2, 'Warning: Could not open "' + p^ + '" for reading, skipping.');
      end
    else
      HandleStream(InputStream);
    Inc(i);
  until (i = FileNames^.Count);
end;

procedure TPasDoc.PrintHelp;
begin
  PrintLn(1, '');
  PrintLn(1, 'General switches');
  PrintLn(1, ' -H, --help                   Show this screen and terminate');
  PrintLn(1, ' -V, --version                Show program version and terminate');
  PrintLn(1, ' -v, --verbosity LEVEL        Set verbosity to LEVEL (default 2)');
  PrintLn(1, 'Parser switches');
  PrintLn(1, ' -c, --descrfile FILE         Read descriptions from FILE');
  PrintLn(1, ' -d, --directive DIRECTIVE    Add a directive');
  PrintLn(1, ' -f, --directivefile FILE     Add directives from FILE');
  PrintLn(1, ' -p, --includeprivate         Include private fields/methods/properties');
  PrintLn(1, ' -s, --sourcefilenames FILE   Read file names from FILE');
  PrintLn(1, ' -u, --includefilepath DIR    Search for include files in DIR');
  PrintLn(1, 'Documentation switches');
  PrintLn(1, ' Formats');
  PrintLn(1, '  -h, --html                  HTML output (default)');
  PrintLn(1, '  -htmlheader FILE            Include FILE as header in HTML output');
  PrintLn(1, '  -htmlfooter FILE            Include File as footer in HTML output');
  PrintLn(1, '  -l, --latex                 LaTeX output');
  PrintLn(1, ' Languages');
  PrintLn(1, '  -b, --brasilian             Brasilian Portuguese');
  PrintLn(1, '  -a, --catalan               Catalan');
  PrintLn(1, '  -m, --dutch                 Dutch');
  PrintLn(1, '  -e, --english               English (default)');
  PrintLn(1, '  -r, --french                French');
  PrintLn(1, '  -g, --german                German');
  PrintLn(1, '  -i, --spanish               Spanish');
  PrintLn(1, ' Other');
  PrintLn(1, '  -j, --notexheaders          No headers and footers in LaTeX output');
  PrintLn(1, '  -k, --keeplinefeeds         Keep description line feeds as found in source');
  PrintLn(1, '  -n, --nohomepage            No pasdoc homepage link in HTML output');
  PrintLn(1, '  -o, --outputdirectory DIR   Write doc file(s) to directory DIR');
  PrintLn(1, '');
  PrintLn(1, 'Visit the rjPasDoc homepage: ' + APP_HOMEPAGE);
end;

procedure TPasDoc.PrintUsage;
begin
  PrintLn(1, APP_FIRST_LINE);
  PrintLn(1, 'Usage: ' + ParamStr(0) + ' {<option(s)>} <FILEs>');
  PrintLn(1, 'Documentation generator for Pascal unit source code files.');
end;

procedure TPasDoc.RemoveExcludedItems(var c: PItemCollection);
var
  i: LongInt;
  p: PItem;
begin
  if (not Assigned(c)) then Exit;
  i := 0;
  while (i < c^.Count) do
    begin
      p := c^.At(i);
      if Assigned(p) and Assigned(p^.Description) and
        //(p^.Description^.FindTag ('EXCLUDE', Offs1, Offs2, Offs3)) then
      (p^.Description^.PosCI('@EXCLUDE') >= 0) then
        begin
          PrintLn(3, 'Hint: Excluding item ' + p^.Name);
          c^.AtFree(i);
        end
      else
        begin
          { P has no excluded tag; but if it is a class, interface, object or
            unit, one of its parts may be excluded }
          if (TypeOf(p^) = TypeOf(TCIO)) then
            begin
              RemoveExcludedItems(PCIO(p)^.Fields);
              RemoveExcludedItems(PItemCollection(PCIO(p)^.Properties));
              RemoveExcludedItems(PItemCollection(PCIO(p)^.Methods));
            end
          else
            begin
              if (TypeOf(p^) = TypeOf(TUnit)) then
                begin
                  RemoveExcludedItems(PUnit(p)^.CIO);
                  RemoveExcludedItems(PUnit(p)^.Constants);
                  RemoveExcludedItems(PItemCollection(PUnit(p)^.FuncsProcs));
                  RemoveExcludedItems(PUnit(p)^.Types);
                  RemoveExcludedItems(PUnit(p)^.Variables);
                end;
            end;
          Inc(i);
        end;
    end;
  if (c^.Count = 0) then
    begin
      Dispose(c, Done);
      c := nil;
    end;
end;

procedure TPasDoc.Run;
var
  Generator: PDocGenerator;
  t1, t2: Cardinal;
begin
  t1 := GetTickCount;
  HandleParams;
  ParseFiles;
  {P^.SearchDescrFileTags(P^.Units);}
  RemoveExcludedItems(Units);
  { check if parsing was successful }
  if (not Assigned(Units)) or (Units^.Count < 1) then
    begin
      PrintLn(1, 'Error: At least one unit must have been successfully parsed to write docs.');
      ReadLn;
      Halt(1);
    end;
  PrintLn(3, 'Creating ' + OUTPUT_FORMAT_NAMES[OutputFormat] + ' documentation file(s)...');
  { create desired output generator }
  case OutputFormat of
    OUTPUT_HTML:
      begin
        Generator := New(PHTMLDocGenerator, Init);
      end;
    OUTPUT_LATEX:
      begin
        Generator := New(PTexDocGenerator, Init);
        PTexDocGenerator(Generator)^.NoTexHeadersAndFooters := NoTexHeadersAndFooters;
      end;
  else
    begin
      Generator := nil; { to prevent 'uninitialized' warnings }
      PrintLn(1, 'Error: Output format unknown.');
      Halt(1);
    end;
  end;
  { copy some fields from P to Generator }
  Generator^.Units := Units;
  Generator^.KeepLineFeeds := KeepLineFeeds;
  Generator^.HtmlHeader := HtmlHeader;
  Generator^.HtmlFooter := HtmlFooter;
  Generator^.Language := Language;
  Generator^.DestDir := DestDir;
  Generator^.NoGeneratorInfo := NoGeneratorInfo;
  Generator^.Units := Units;
  Generator^.BuildLinks;
  Generator^.ExpandDescriptions;
  Generator^.LoadDescriptionFiles(DescriptionFiles);
  Generator^.WriteDocumentation;
  Dispose(Generator, Done);

  t2 := GetTickCount;
  PrintLn(2, 'Done, worked ' + IntToStr((t2 - t1) div 1000) + ' second(s).');
end;

procedure TPasDoc.SearchDescrFileTags(var c: PItemCollection);
var
  Found: Boolean;
  i: LongInt;
  Offs1: LongInt;
  Offs2: LongInt;
  Offs3: LongInt;
  p: PItem;
  s: string;
begin
  if (not Assigned(c)) then Exit;
  i := 0;
  while (i < c^.Count) do
    begin
      p := c^.At(i);
      Inc(i);
      if (not Assigned(p)) then Continue;
      if Assigned(p^.Description) then
        begin
          Offs1 := 0;
          repeat
            Found := p^.Description^.FindTag('DESCRFILE', Offs1, Offs2, Offs3);
            if Found then
              begin
                p^.Description^.ExtractTag(Offs1, Offs2, Offs3, s);
                PrintLn(3, 'Adding description file "' + s + '"');
                AddDescriptionFile(s);
                Offs1 := Offs3 + 1;
              end;
          until (not Found);
        end;
      if (TypeOf(p^) = TypeOf(TCIO)) then
        begin
          SearchDescrFileTags(PCIO(p)^.Fields);
          SearchDescrFileTags(PItemCollection(PCIO(p)^.Methods));
          SearchDescrFileTags(PItemCollection(PCIO(p)^.Properties));
        end;
      if (TypeOf(p^) = TypeOf(TUnit)) then
        begin
          SearchDescrFileTags(PUnit(p)^.CIO);
          SearchDescrFileTags(PUnit(p)^.Constants);
          SearchDescrFileTags(PItemCollection(PUnit(p)^.FuncsProcs));
          SearchDescrFileTags(PUnit(p)^.Types);
          SearchDescrFileTags(PUnit(p)^.Variables);
        end;
    end;
end;

procedure TPasDoc.SetLanguage(const NewLanguage: TLanguage);
begin
  if LanguageInitialized and (Language <> NewLanguage) then
    begin
      PrintLn(1, 'Error - You can only define one language.');
      Halt(1);
    end;
  Language := NewLanguage;
  LanguageInitialized := True;
end;

end.

