{
@abstract(provides LaTeX doc generator object)
@author(Michael van Canneyt (michael@tfdec1.fys.kuleuven.ac.be))
@author(Marco Schmidt (marcoschmidt@geocities.com))
@created(9 Sep 1998)
@lastmod(10 Apr 2000)

Implements an object to generate TeX documentation, overriding many of
@link(TDocGenerator)'s virtual methods.
TeX is a free text setting package that produces output well-suited for
printing.
}

unit tex;

{$I platform.inc}

interface

uses
  AppInfo, arrays, FileStre, GenDoc, Items, Objects, Parsing, streams, Texts,
  Time;

const
  { name of the TeX output file (there is just a single file for all the
    documentation) }
  TEX_OUTPUT_FILENAME = 'docs';

type
  { pointer to @link(TTexDocGenerator) }
  PTexDocGenerator = ^TTexDocGenerator;
  { @abstract(generates TeX documentation)
    Extends @link(TDocGenerator) and overwrites many of its methods to
    generate output in TeX format.
    This type of output is well suited to be printed. }
  TTexDocGenerator = object (TDocGenerator)
    { number of cells we've already written in current table row }
    CellCounter: LongInt;
    { default is false; if true pasdoc will not write document type at
      the beginning or end-of-document information at the end so that
      one can include the .tex output file in any other tex document }
    NoTexHeadersAndFooters: Boolean;
    { number of cells (= columns) per table row }
    NumCells: LongInt;
    { Implements @link(TDocGenerator.ConvertChar).
      Returns a string that can be used to output character C in TeX.
      Several special characters must be escaped, e.g. the backslash. }
    function ConvertChar (c: Char): string; virtual;
    { Returns a link to an anchor within a document. }
    function CreateLink (Item: PItem): string; virtual;
    function CreateReferencedLink (ItemName, Link: string): string; virtual;
    { Returns (La)TeX file extension '.tex'. }
    function GetFileExtension: string; virtual;
    { Writes an anchor, i.e., the name of the label by which item I will be
      referenced in other item's comments.
      Will be constructed from @link(TItem.AnchorNumber) and
      @link(AnchorToString). }
    procedure WriteAnchor (i: PItem);
    { Writes all information of a single class, interface or object (CIO)
      to output, starting at heading level HL. }
    procedure WriteCIO (HL: Byte; var CIO: TCIO); virtual;
    { Writes information on all classes, interfaces and objects from C to
      output, starting at heading level HL. }
    procedure WriteCIOs (HL: Byte; c: PItemCollection); virtual;
    { Writes a comment to output file. }
    procedure WriteComment (const s: string); virtual;
    { Writes all documentation files - number will vary between output
      formats. }
    procedure WriteDocumentation; virtual;
    { original comment }
    procedure WriteFields (HL: Byte; c: PItemCollection);
    { Writes functions and procedures / methods from FuncsProcs to the docs,
      at heading level HL. If methods is true, it's a set of methods,
      otherwise a set of functions and procedures. }
    procedure WriteFuncsProcs (HL: Byte; Methods: Boolean; FuncsProcs:
      PMethodCollection); virtual;
    { Writes a heading S at heading level HL.
      HL should be from 1 to 4.
      The smaller the number, the larger will be the resulting heading. }
    procedure WriteHeading (HL: Byte; const s: string); virtual;
    { Writes descriptions for all items in the given collection (e.g.
      variables, constants etc.) at heading level HL with given heading. }
    procedure WriteItems (HL: Byte; Heading: string; i: PItemCollection); virtual;
    procedure WriteMethodOverview (p: PMethodCollection); virtual;
    procedure WriteMethodDeclaration (const s: string); virtual;
    procedure WriteText (t: PText); virtual;
    procedure WriteProperties (HL: Byte; p: PPropertyCollection); virtual;
    { Writes necessary introductory statements for a valid LaTeX document,
      includes several packages (e.g. for tables). }
    procedure WriteStartOfDocument;
    { Writes all documentation for one unit. }
    procedure WriteUnit (HL: Byte; U: PUnit); virtual;
    procedure WriteUnitDescription (HL: Byte; d: PText); virtual;
  private
    {  }
    procedure WriteEndOfAnchor (const Name: string);
    {  }
    procedure WriteEndOfDocument;
    procedure WriteEndOfLink (const Name: string); virtual;
    procedure WriteEndOfTable2Columns;
    procedure WriteEndOfTableCell;
    procedure WriteEndOfTableRow;
    procedure WriteStartOfLink (const Name: string);
    procedure WriteStartOfTable1Column (t: string);
    procedure WriteStartOfTable2Columns (t1, t2: string);
    procedure WriteStartOfTable3Columns (t1, t2, T3: string);
    procedure WriteStartOfTableRow;
  end;

implementation

uses
  Msg;

function TTexDocGenerator.ConvertChar (c: Char): string;
begin
  if (c in ['&', '^', '_', '$', '%', '#', '{', '}'])
    then ConvertChar := '\' + c
  else
    if (c in ['<', '>'])
      then ConvertChar := '$' + c + '$'
    else
      begin
        if (c = '\')
          then ConvertChar := '\verb|\|'
        else ConvertChar := c;
      end;
end;

function TTexDocGenerator.CreateLink (Item: PItem): string;
begin
  CreateLink := AnchorToString (Item^.AnchorNumber);
end;

function TTexDocGenerator.CreateReferencedLink (ItemName, Link: string): string;
begin
  CreateReferencedLink := '{\ttfamily ' + ConvertString (ItemName) +
    '} (p. \pageref{' + Link + '})';
end;

function TTexDocGenerator.GetFileExtension: string;
begin
  GetFileExtension := '.tex';
end;

procedure TTexDocGenerator.WriteCIO (HL: Byte; var CIO: TCIO);
var
  s                 : string;
begin
  case CIO.MyType of
    CIO_CLASS: s := getString (TEXT_CLASS);
    CIO_INTERFACE: s := getString (TEXT_INTERFACE);
    CIO_OBJECT: s := getString (TEXT_OBJECT);
  else
    begin
      PrintLn (1, 'Fatal error in TTexDocGenerator.WriteCIO: unknown CIO type.');
      Halt (1);
    end;
  end;
  s := s + ' ' + CIO.Name;
  WriteAnchor (@CIO);
  WriteDescription (HL + 1, s, CIO);
  WriteFields (HL + 1, CIO.Fields);
  WriteFuncsProcs (HL + 1, True, CIO.Methods);
  WriteProperties (HL + 1, CIO.Properties);
end;

procedure TTexDocGenerator.WriteCIOs (HL: Byte; c: PItemCollection);
var
  i                 : LongInt;
  Item              : PCIO;
begin
  if (not Assigned (c)) or (c^.Count < 1)
    then Exit;
  PrintLn (4, 'Writing docs for classes/interfaces/objects');
  for i := 0 to c^.Count - 1 do
    begin
      Item := c^.At (i);
      { note that we do _not_ increase the heading level
        we might otherwise run out of heading levels for TeX etc. }
      if Assigned (Item)
        then WriteCIO (HL, Item^);
    end;
end;

procedure TTexDocGenerator.WriteComment (const s: string);
begin
  Stream^.WriteChar ('%');
  Stream^.WriteLine (s)
end;

procedure TTexDocGenerator.WriteDocumentation;
begin
  if (not CreateStream (TEX_OUTPUT_FILENAME)) then
    begin
      PrintLn (1, 'Error - could not create output stream.');
      Exit;
    end;
  WriteComment (APP_FIRST_LINE);
  WriteComment (getString (TEXT_WARNING_AUTOMATIC));
  WriteStartOfDocument;
  WriteUnits (1);
  WriteEndOfDocument;
  CloseStream;
end;

procedure TTexDocGenerator.WriteEndOfDocument;
begin
  if NoTexHeadersAndFooters
    then Exit;
  PrintLn (4, 'Writing end of TeX document.');
  if (not Assigned (Stream)) then
    begin
      PrintLn (2, 'Error TTexDocGenerator.WriteEndOfDocument: No stream available.');
      Exit;
    end;
  Stream^.WriteLine ('\end{document}');
end;

procedure TTexDocGenerator.WriteEndOfAnchor (const Name: string);
begin
end;

procedure TTexDocGenerator.WriteEndOfLink (const Name: string);
begin
end;

(*
procedure TTexDocGenerator.WriteEndOfTable1Column;
begin
  Stream^.WriteLine('\hline');
  Stream^.WriteLine('\end{tabular}');
end;*)

procedure TTexDocGenerator.WriteEndOfTable2Columns;
begin
  { If you insert again the \hline at the endofrow,
    then remove this one }
  Stream^.WriteLine ('\hline');
  Stream^.WriteLine ('\end{tabularx}');
end;

(*procedure TTexDocGenerator.WriteEndOfTable3Columns;
begin
        { If you insert again the \hline at the endofrow,
          then remove this one }
  Stream^.WriteLine('\hline');
  Stream^.WriteLine('\end{tabularx}');
end;*)

procedure TTexDocGenerator.WriteEndOfTableCell;
begin
  Inc (CellCounter);
  if (CellCounter < NumCells)
    then Stream^.WriteString (' & ');
end;

procedure TTexDocGenerator.WriteEndOfTableRow;
begin
  Stream^.WriteLine (' \\');
  {  Stream^.WriteLine('\hline');}
end;

procedure TTexDocGenerator.WriteFields (HL: Byte; c: PItemCollection);
begin
  WriteItems (HL, getString (TEXT_FIELDS), c);
end;

procedure TTexDocGenerator.WriteFuncsProcs (HL: Byte; Methods: Boolean;
  FuncsProcs: PMethodCollection);
var
  i                 : LongInt;
  m                 : PMethod;
begin
  if (not Assigned (FuncsProcs)) or (FuncsProcs^.Count < 1)
    then Exit;
  PrintLn (4, 'Writing funcs/procs/methods.');
  if Methods
    then WriteHeading (HL, getString (TEXT_METHODS))
  else WriteHeading (HL, getString (TEXT_FUNCTIONS_AND_PROCEDURES));
  for i := 0 to FuncsProcs^.Count - 1 do
    begin
      m := FuncsProcs^.At (i);
      { write a label that can be referenced elsewhere in the output file }
      WriteAnchor (m);
      { write method declaration in courier font, then line feed }
      Stream^.WriteString ('{\ttfamily ');
      WriteString (m^.FullDeclaration);
      Stream^.WriteLine ('}\\');
      { write description of method }
      WriteText (m^.GetDescription);
      { write line feed in output so that next declaration will be at
        the beginning of a new line, with one line distance }
      Stream^.WriteLine ('\\');
      Stream^.WriteLine ('\\');
    end;
end;

procedure TTexDocGenerator.WriteHeading (HL: Byte; const s: string);
begin
  if (not Assigned (Stream)) then
    begin
      PrintLn (2, 'Error TTexDocGenerator.WriteHeading: No stream available.');
      Exit;
    end;
  case HL of
    1: Stream^.WriteString ('\chapter*{');
    2: Stream^.WriteString ('\section*{');
    3: Stream^.WriteString ('\subsection*{');
    4: Stream^.WriteString ('\subsubsection*{');
  else Stream^.WriteChar ('{');
  end;
  WriteString (s);
  Stream^.WriteLine ('}');
end;

procedure TTexDocGenerator.WriteItems (HL: Byte; Heading: string;
  i: PItemCollection);
var
  j                 : LongInt;
  Item              : PItem;
begin
  if (not Assigned (i)) or (i^.Count < 1)
    then Exit;
  PrintLn (4, 'Writing items (' + Heading + ')');
  WriteHeading (HL, Heading);
  for j := 0 to i^.Count - 1 do
    begin
      Item := i^.At (j);
      WriteAnchor (Item);
      Stream^.WriteString ('\noindent ');
      Stream^.WriteString ('{\ttfamily ');
      WriteString (Item^.Name);
      Stream^.WriteString ('}');
      Stream^.WriteLine ('\\');
      WriteText (Item^.GetDescription);
      Stream^.WriteLine ('\\');
    end;
end;

procedure TTexDocGenerator.WriteMethodOverview (p: PMethodCollection);
var
  i                 : LongInt;
  Item              : PMethod;
begin
  if (not Assigned (p)) or (p^.Count < 1)
    then Exit;
  WriteStartOfTable2Columns ('Name', 'Page');
  for i := 0 to p^.Count - 1 do
    begin
      Item := p^.At (i);
      WriteStartOfTableRow;
      WriteMethodDeclaration (Item^.FullDeclaration);
      WriteEndOfTableCell;
      WriteStartOfLink (Item^.FullLink);
      WriteEndOfLink (Item^.FullLink);
      WriteEndOfTableCell;
      WriteEndOfTableRow;
    end;
  WriteEndOfTable2Columns;
end;

procedure TTexDocGenerator.WriteMethodDeclaration (const s: string);
begin
  { the \ttfamily jumps to teletype font.}
  Stream^.WriteString ('{\ttfamily ');
  WriteString (s);
  Stream^.WriteLine ('} \par');
end;

procedure TTexDocGenerator.WriteProperties (HL: Byte; p: PPropertyCollection);
begin
end;

(*
Procedure TTexDocGenerator.WritePropertyOverview(P: PPropertyCollection);
Var
  I: LongInt;
  Item: PProperty;
begin
  if (not Assigned(P)) or (P^.Count < 1)
  then exit;
  WriteStartOfTable2Columns('Name','Page');
  for I := 0 to P^.Count - 1 do
  begin
    Item := P^.At(I);
    WriteStartOfTableRow;
    WritePropertyDeclaration(Item);
    WriteEndOfTableCell;
    WriteStartOfLink(Item^.FullLink);
    WriteEndOfLink(Item^.FullLink);
    WriteEndOfTableCell;
    WriteEndOfTableRow;
  end;
  WriteEndOfTable2Columns;
end;

procedure TTexDocGenerator.WritePropertyDeclaration(P : PProperty);
begin
  Stream^.WriteString('{\ttfamily ');
  WriteString(p^.GetDeclaration);
  Stream^.WriteLine('} \par');
end;
*)

procedure TTexDocGenerator.WriteAnchor (i: PItem);
begin
  Stream^.WriteString ('\label{' + AnchorToString (i^.AnchorNumber) + '}');
end;

procedure TTexDocGenerator.WriteStartOfDocument;
begin
  if NoTexHeadersAndFooters
    then Exit;
  PrintLn (4, 'Writing start of TeX document.');
  if (not Assigned (Stream)) then
    begin
      PrintLn (2, 'Error in TTexDocGenerator.WriteStartOfDocument: No stream available.');
      Exit;
    end;
  Stream^.WriteLine ('\documentclass{report}');
  Stream^.WriteLine ('\usepackage{tabularx}');
  Stream^.WriteLine ('\begin{document}');
end;

procedure TTexDocGenerator.WriteStartOfLink (const Name: string);
begin
  Stream^.WriteString ('(\pageref{' + Name + '})');
end;

procedure TTexDocGenerator.WriteStartOfTable1Column (t: string);
begin
  Stream^.WriteLine ('\begin{tabular}{|l|}');
  Stream^.WriteLine ('\hline');
  NumCells := 1;
end;

procedure TTexDocGenerator.WriteStartOfTable2Columns (t1, t2: string);
begin
  Stream^.WriteLine ('\begin{tabularx}{\linewidth}{|lX|}');
  Stream^.WriteLine ('\hline');
  WriteString (t1);
  Stream^.WriteString (' & ');
  WriteString (t2);
  Stream^.WriteLine (' \\');
  Stream^.WriteLine ('\hline');
  NumCells := 2;
end;

procedure TTexDocGenerator.WriteStartOfTable3Columns (t1, t2, T3: string);
begin
  Stream^.WriteLine ('\begin{tabularx}{\linewidth}{|llX|}');
  Stream^.WriteLine ('\hline');
  WriteString (t1);
  Stream^.WriteString (' & ');
  WriteString (t2);
  Stream^.WriteString (' & ');
  WriteString (T3);
  Stream^.WriteLine (' \\');
  Stream^.WriteLine ('\hline');
  NumCells := 3;
end;

procedure TTexDocGenerator.WriteStartOfTableRow;
begin
  CellCounter := 0;
end;

procedure TTexDocGenerator.WriteText (t: PText);
var
  c                 : Char;
  i                 : LongInt;
  l                 : LongInt;
begin
  if (t = nil) or (t^.Content < 1)
    then Exit;
  l := 0;
  for i := 0 to t^.Content - 1 do
    begin
      c := t^.Data[i];
      if (c = #9) or (c = #10) or (c = #13)
        then c := ' ';
      Stream^.WriteChar (c);
      Inc (l);
      if (l > 80) and (t^.Data[i] = ' ')
        then
        begin
          Stream^.WriteLine ('');
          l := 0;
        end;
    end;
  {  Stream^.WriteLine('\\');}
end;

procedure TTexDocGenerator.WriteUnit (HL: Byte; U: PUnit);
begin
  if (not Assigned (U))
    then Exit;
  PrintLn (3, 'Writing docs for unit ' + U^.Name);
  {  if (Assigned(U^.DetailedDescription))
    then D := U^.DetailedDescription
    else D := U^.Description;
    WriteUnitDescription(HL, D);}
  WriteDescription (HL, getString (TEXT_UNIT) + ' ' + U^.Name, U^);
  WriteConstants (HL + 1, U^.Constants);
  WriteVariables (HL + 1, U^.Variables);
  WriteTypes (HL + 1, U^.Types);
  WriteFuncsProcs (HL + 1, False, U^.FuncsProcs);
  WriteCIOs (HL + 1, U^.CIO);
end;

procedure TTexDocGenerator.WriteUnitDescription (HL: Byte; d: PText);
var
  i                 : LongInt;
begin
  if (not Assigned (d)) or (d^.Content < 1) or (not Assigned (d^.Data))
    then Exit;
  PrintLn (4, 'Writing unit description.');
  WriteHeading (HL, getString (TEXT_DESCRIPTION));
  for i := 0 to d^.Content - 1 do
    begin
      Stream^.WriteChar (d^.Data[i]);
    end;
end;

begin
end.

